/*---------------------------------------------------------------------------*
Project: NitroWiFi - shared - demos
File: wcm_control.c

Copyright 2005.2006 Nintendo.  All rights reserved.

These coded instructions, statements, and computer programs contain
proprietary information of Nintendo of America Inc. and/or Nintendo
Company Ltd., and are protected by Federal copyright law.  They may
not be disclosed to third parties or copied or duplicated in any form,
in whole or in part, without the prior written consent of Nintendo.

$Log: wcm_control.c,v $
Revision 1.13  03/10/2006 09:22:42  kitase_hirotake
  INDENT SOURCE

Revision 1.12  2005/11/02 04:13:42  terui
Fix comment.

Revision 1.11  09/13/2005 13:14:14  terui
revisions in conjunction with adding the DisconnectAP function

Revision 1.10  09/08/2005 10:42:38  adachi_hiroaki
Made it possible to get the number of round scans. Also made changes such that debug display within the callback function only appears once.

Revision 1.9  09/06/2005 02:53:44  yasu
revisions in conjunction with moving nitroWiFi to the env.h NitroSDK

Revision 1.8  09/05/2005 04:00:49  terui
Revised such that scan method and power management mode can be changed externally.

Revision 1.7  2005/08/29 06:39:02  yosizaki
add SetWcmManualConnect(), ConnectAP().

Revision 1.6  08/23/2005 01:56:35  seiki_masashi
rename ENV_GetBinaryAndLength() to ENV_GetBinaryAndSize()

Revision 1.5  08/22/2005 12:10:19  yasu
revised for reference to SiteDefs

Revision 1.4  07/27/2005 08:36:22  terui
Add SetWcmThreadPriority API

Revision 1.3  07/22/2005 11:18:56  terui
unified denotation when specifying stack bottom

Revision 1.2  07/22/2005 10:36:17  terui
Fixed bug in specifying position of the bottom of the thread stack.

Revision 1.1  07/21/2005 08:38:16  terui
Added new source files common to samples

$NoKeywords: $
 *---------------------------------------------------------------------------*/
#include <nitro/os.h>
#include <nitroWiFi.h>
#include "wcm_control.h"
#include "ap_info.h"

/*---------------------------------------------------------------------------*
constant definitions
 *---------------------------------------------------------------------------*/
#define LOCAL_WCM_THREAD_STACK_SIZE     1024    // Allocates a large stack (can be changed)
#define LOCAL_WCM_THREAD_PRIORITY       24      // a low priority is sufficient (can be changed)
#define LOCAL_WCM_DMA_NO                3       // DMA number used in wireless functionality (can be changed)
#define LOCAL_WCM_APLIST_COUNT_MAX      22      // the number of AP data lists to be managed within the WCM library
#define LOCAL_WCM_APLIST_BUF_SIZE       ((LOCAL_WCM_APLIST_COUNT_MAX * WCM_APLIST_BLOCK_SIZE) + 12)
#define LOCAL_WCM_THREAD_MSGQ_CAPACITY  3       // the number of levels in the message queue
#define LOCAL_WCM_MSG_TERMINATE         ((void*)0)
#define LOCAL_WCM_MSG_FOUND_AP          ((void*)1)
#define LOCAL_WCM_MSG_DISCONNECTED      ((void*)2)
#define LOCAL_WCM_MSG_ERROR             ((void*)3)
#define LOCAL_WCM_MSG_CONNECT_AP        ((void*)4)
#define LOCAL_WCM_MSG_DISCONNECT        ((void*)5)

// string constant for displaying the WCM library operations log
static const char*  gWcmResultText[8] =
{
    { "SUCCESS" } ,         { "FAILURE" } ,
    { "PROGRESS" } ,        { "ACCEPT" } ,
    { "REJECT" } ,          { "WMDISABLE" } ,
    { "NOT_ENOUGH_MEM" } ,  { "FATAL_ERROR" }
};
static const char*  gWcmNotifyText[10] =
{
    { "COMMON" } ,          { "STARTUP" } ,
    { "CLEANUP" } ,         { "BEGIN_SEARCH" } ,
    { "END_SEARCH" } ,      { "CONNECT" } ,
    { "DISCONNECT" } ,      { "FOUND_AP" } ,
    { "SEARCH_AROUND" } ,   { "TERMINATE" }
};

// debugging string display settings
#define LOCAL_DEBUG_PRINT   1
#if LOCAL_DEBUG_PRINT
#define DebugPrintf OS_TPrintf
#else
#define DebugPrintf(...)    ((void)0)
#endif

/*---------------------------------------------------------------------------*
internal variable definitions
 *---------------------------------------------------------------------------*/

// for thread controls
static OSThread         gWcmThread;
static u32          gWcmStack[LOCAL_WCM_THREAD_STACK_SIZE / sizeof( u32 ) ];
static OSMessageQueue   gWcmMessageQ;
static OSMessage        gWcmMsgArray[LOCAL_WCM_THREAD_MSGQ_CAPACITY];
static OSThreadQueue    gWcmThreadQ;

// for WCM library controls
static volatile BOOL    gWcmTerminateFlag;
static vs32         gWcmResult;
static void*        gWcmBuf = NULL;
static WCMConfig    gWcmConfig;
static WMBssDesc gWcmBssDesc    ATTRIBUTE_ALIGN(4);
static u32  gWcmScanType = WCM_OPTION_SCANTYPE_ACTIVE;
static u32  gWcmPowerMode = WCM_OPTION_POWER_ACTIVE;

static BOOL gWcmManualConnect = FALSE;
static u32  gWcmSearchAroundCount = 0;

static const WcmControlApInfo*  constApInfoPtr = NULL;

/*---------------------------------------------------------------------------*
internal function prototype
 *---------------------------------------------------------------------------*/
static void ControlThread(void* arg);
static void FreezeThread(void);
static s32  CallFunction(s16 notifyId);
static void ControlCallback(WCMNotice* arg);

/*---------------------------------------------------------------------------*
Name: SetWcmScanType

Description: Sets the scan method for searching for APs.
The default is the active scan method.
Need to call before InitWcmControl().

Arguments: type    -   Specifies the WCM_OPTION_SCANTYPE_* constant.

Returns: None.
 *---------------------------------------------------------------------------*/
void SetWcmScanType(u32 type)
{
    switch (type)
    {
    case WCM_OPTION_SCANTYPE_PASSIVE:
        gWcmScanType = WCM_OPTION_SCANTYPE_PASSIVE;
        break;

    default:
        gWcmScanType = WCM_OPTION_SCANTYPE_ACTIVE;
    }
}

/*---------------------------------------------------------------------------*
Name: SetWcmPowerMode

Description: Specifies the power management method when communicating with an AP.
The default is the ordinary active method.
Need to call before InitWcmControl().

Arguments: mode    -   Specifies the WCM_OPTION_POWER_* constant.

Returns: None.
 *---------------------------------------------------------------------------*/
void SetWcmPowerMode(u32 mode)
{
    switch (mode)
    {
    case WCM_OPTION_POWER_SAVE:
        gWcmPowerMode = WCM_OPTION_POWER_SAVE;
        break;

    default:
        gWcmPowerMode = WCM_OPTION_POWER_ACTIVE;
    }
}

/*---------------------------------------------------------------------------*
Name: SetWcmManualConnectMode

Description: Settings to select a connection manually from the AP list.
Default setting is FALSE (auto-connection).
Need to call before InitWcmControl().

Arguments: enable -   TRUE if manual connection.

Returns: None.
 *---------------------------------------------------------------------------*/
void SetWcmManualConnect(BOOL enable)
{
    gWcmManualConnect = enable;
}

/*---------------------------------------------------------------------------*
Name: ConnectAP

Description: connects to the specified BssDesc AP

Arguments: bd     -   Pointer to the structure indicated by the connection AP.

Returns: None.
 *---------------------------------------------------------------------------*/
void ConnectAP(const WMBssDesc* pbd)
{
    SDK_NULL_ASSERT(pbd);
    MI_CpuCopy8(pbd, &gWcmBssDesc, /*sizeof(WMBssDesc)*/ WCM_APLIST_SIZE);
    (void)OS_SendMessage(&gWcmMessageQ, LOCAL_WCM_MSG_CONNECT_AP, OS_MESSAGE_NOBLOCK);
}

/*---------------------------------------------------------------------------*
Name: DisconnectAP

Description: attempts to disconnect from the AP connection

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
void DisconnectAP(void)
{
    OSIntrMode  e = OS_DisableInterrupts();
    if (WCM_GetPhase() == WCM_PHASE_DCF)
    {
        (void)OS_SendMessage(&gWcmMessageQ, LOCAL_WCM_MSG_DISCONNECT, OS_MESSAGE_NOBLOCK);
    }
    (void)OS_RestoreInterrupts(e);
}

/*---------------------------------------------------------------------------*
Name: InitWcmControl/InitWcmControlByApInfo

Description: Initializes internal variables and launches the WCM control thread.
Nothing is performed if the thread has already been launched.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
void InitWcmControl(void)
{
    InitWcmControlByApInfo(&constApInfo);
}

void InitWcmControlByApInfo(const WcmControlApInfo* ptr)
{
// Is the thread variable initialized even once?
    if (gWcmThread.id != 0)
    {
// Is the thread in an ended state?
        if (OS_IsThreadTerminated(&gWcmThread) == FALSE)
        {
            return;
        }
    }

    constApInfoPtr = ptr;

// initializes the message queue
    OS_InitMessageQueue(&gWcmMessageQ, gWcmMsgArray, LOCAL_WCM_THREAD_MSGQ_CAPACITY);

// initializes the thread queue
    OS_InitThreadQueue(&gWcmThreadQ);

// initializes the variables
    MI_CpuClearFast(&gWcmBssDesc, sizeof(WMBssDesc));

// creates a new thread using the thread variables
    OS_CreateThread(&gWcmThread, ControlThread, NULL, (void*)(gWcmStack + (LOCAL_WCM_THREAD_STACK_SIZE / sizeof(u32))),
                    LOCAL_WCM_THREAD_STACK_SIZE, LOCAL_WCM_THREAD_PRIORITY);

// launches the created thread
    gWcmTerminateFlag = FALSE;
    OS_WakeupThreadDirect(&gWcmThread);
}

/*---------------------------------------------------------------------------*
Name: TerminateWcmControl

Description: Ends the WCM control thread.
Nothing is performed if the thread has not been launched.
NOTICE: As this function only serves as a trigger to end the thread,
be aware that it will return without waiting for the thread itself to end.
If you want to wait for the thread itself to end, you will need to call
OS_JoinThread after calling OS_SendMessage.
In addition, if the WCM_GetPhase is set to return WCM_PHASE_NULL,
the thread has ended.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
void TerminateWcmControl(void)
{
// Is the thread variable initialized even once?
    if (gWcmThread.id != 0)
    {
// Is the thread in an ended state?
        if (OS_IsThreadTerminated(&gWcmThread) == FALSE)
        {
// Is it already in a state of having accepted an end request?
            if (gWcmTerminateFlag == FALSE)
            {
                gWcmTerminateFlag = TRUE;
                (void)OS_SendMessage(&gWcmMessageQ, LOCAL_WCM_MSG_TERMINATE, OS_MESSAGE_NOBLOCK);
            }
        }
    }
}

/*---------------------------------------------------------------------------*
Name: SetWcmThreadPriority

Description: Changes the priority of the thread performing WCM library state management.

Arguments: priority    -   Specifies a thread priority between 0 and 31.

Returns: u8  -   Returns the priority setting prior to being changed.
Returns 0xff if for some reason the change to priority fails.
 *---------------------------------------------------------------------------*/
u8 SetWcmThreadPriority(u8 priority)
{
    u8  beforePrio = 0xff;

    if (priority < 32)
    {
        if ((gWcmThread.id != 0) && (OS_IsThreadTerminated(&gWcmThread) == FALSE))
        {
            beforePrio = (u8) OS_GetThreadPriority(&gWcmThread);
            if (FALSE == OS_SetThreadPriority(&gWcmThread, (u32) priority))
            {
                beforePrio = 0xff;
            }
        }
    }

    return beforePrio;
}

/*---------------------------------------------------------------------------*
Name: GetWcmSearchAroundCount

Description: Returns the number of times all channels have been scanned.
0 immediately after StartUp.

Arguments: None.

Returns: How many times scanned?
 *---------------------------------------------------------------------------*/
u32 GetWcmSearchAroundCount(void)
{
    return gWcmSearchAroundCount;
}

/*---------------------------------------------------------------------------*
Name: ControlThread

Description: Thread performing WCM library state management.
While the gWcmTerminateFlag variable is FALSE attempts will be made to connect to the specified AP
or to maintain that connection if a connection has already been made.
When the TerminateWcmControl function is called, tge gWcmTerminateFlag variable
is changed to true, and subsequently the WCM library will end and the thread itself
will attempt to end.
In addition, if a FATAL ERROR occurs within the WCM library,
this thread will internally enter an infinite loop.

Arguments: arg     -   Not used.

Returns: None.
 *---------------------------------------------------------------------------*/
static void ControlThread(void* arg)
{
#pragma unused(arg)

    void*       wcmBuf;
    void*       wcmAplistBuf;
    s32         wcmResult;
    OSMessage   tempMessage;

// blocks off the memory to be used by the WCM library
    wcmBuf = OS_Alloc(WCM_WORK_SIZE + LOCAL_WCM_APLIST_BUF_SIZE);

    if (wcmBuf == NULL)
    {
        /* if the memory is unavailable, subsequent operations will fail, so end */
        return;
    }

    wcmAplistBuf = (void*)((u32) wcmBuf + WCM_WORK_SIZE);

// initializes the WCM library
    wcmResult = WCM_Init(wcmBuf, WCM_WORK_SIZE);
    DebugPrintf("WCM_Init() -> %s\n", gWcmResultText[wcmResult]);
    if (wcmResult != WCM_RESULT_SUCCESS)
    {
        /*
* If an error is returned here, it is highly possible that use of the library is in error.
* It is possible that other threads have already initialized the WCM library.
         */
        OS_Free(wcmBuf);
        return;
    }

// WCM library operation settings
    gWcmConfig.dmano = LOCAL_WCM_DMA_NO;
    gWcmConfig.pbdbuffer = wcmAplistBuf;
    gWcmConfig.nbdbuffer = LOCAL_WCM_APLIST_BUF_SIZE;
    gWcmConfig.nbdmode = WCM_APLIST_MODE_IGNORE;

// connection start
    while (gWcmTerminateFlag == FALSE)
    {
        switch (WCM_GetPhase())
        {
        case WCM_PHASE_WAIT:
// launch wireless functionality
            wcmResult = CallFunction(WCM_NOTIFY_STARTUP);
            break;

        case WCM_PHASE_IDLE:
            if (gWcmBssDesc.length == 0)
            {
// begin AP auto-search
                WCM_ClearApList();
                wcmResult = CallFunction(WCM_NOTIFY_BEGIN_SEARCH);
            }
            else
            {
// connect to AP
                wcmResult = CallFunction(WCM_NOTIFY_CONNECT);
                if (wcmResult == WCM_RESULT_SUCCESS)
                {
// clear here to re-perform search when attempting to connect next
                    MI_CpuClearFast(&gWcmBssDesc, sizeof(WMBssDesc));
                }
            }
            break;

        case WCM_PHASE_SEARCH:
// wait until an AP is detected
            (void)OS_ReceiveMessage(&gWcmMessageQ, &tempMessage, OS_MESSAGE_BLOCK);

            /*
* There are three types of messages that could be received: a WCM stop request, an AP detection
notification or an error notification. If an AP detection notification, the AP detected by this sample is
* the AP to which the connection should be made, and the data is cleared.
             */
            if (tempMessage == LOCAL_WCM_MSG_FOUND_AP)
            {
                /* end search here is if auto-connect mode */
                if (!gWcmManualConnect)
                {
                    (void)WCM_LockApList(WCM_APLIST_LOCK);
                    if (WCM_CountApList() >= 1)
                    {
// Only clears the data for the first found AP.
                        MI_CpuCopyFast(WCM_PointApList(0), &gWcmBssDesc, WCM_APLIST_SIZE);
                        DebugPrintf("  BSS-ID: %02x-%02x-%02x-%02x-%02x-%02x\n", gWcmBssDesc.bssid[0],
                                    gWcmBssDesc.bssid[1], gWcmBssDesc.bssid[2], gWcmBssDesc.bssid[3],
                                    gWcmBssDesc.bssid[4], gWcmBssDesc.bssid[5]);
                        DebugPrintf("  ESS-ID: %s\n", gWcmBssDesc.ssid);
                    }
                    (void)WCM_LockApList(WCM_APLIST_UNLOCK);
                    wcmResult = CallFunction(WCM_NOTIFY_END_SEARCH);
                }
            }

            /* end the search if there is a connection request */
            else if (tempMessage == LOCAL_WCM_MSG_CONNECT_AP)
            {
                wcmResult = CallFunction(WCM_NOTIFY_END_SEARCH);
            }
            break;

        case WCM_PHASE_DCF:
// wait until disconnect notification
            (void)OS_ReceiveMessage(&gWcmMessageQ, &tempMessage, OS_MESSAGE_BLOCK);

            /*
* There are three types of messages that could be received: a WCM stop request, an AP disconnect
notification or an error notification. No process is to be performed regardless of the message received.
             */
            if (tempMessage == LOCAL_WCM_MSG_DISCONNECT)
            {
                wcmResult = CallFunction(WCM_NOTIFY_DISCONNECT);
            }
            break;

        case WCM_PHASE_IRREGULAR:
// wait a little, as it is highly possible that there has been a FIFO overflow
            OS_Sleep(1);

// a forced stop is the only enabled process during the IRREGULAR phase
            wcmResult = CallFunction(WCM_NOTIFY_TERMINATE);
            break;

        case WCM_PHASE_FATAL_ERROR:
        default:
            FreezeThread();
        }
    }

// end process
    while (gWcmTerminateFlag != FALSE)
    {
        switch (WCM_GetPhase())
        {
        case WCM_PHASE_WAIT:
// escape from the loop
            gWcmTerminateFlag = FALSE;
            break;

        case WCM_PHASE_IDLE:
// stop wireless functionality
            wcmResult = CallFunction(WCM_NOTIFY_CLEANUP);
            break;

        case WCM_PHASE_SEARCH:
// stop AP auto-search
            wcmResult = CallFunction(WCM_NOTIFY_END_SEARCH);
            break;

        case WCM_PHASE_DCF:
// disconnect from the AP
            wcmResult = CallFunction(WCM_NOTIFY_DISCONNECT);
            break;

        case WCM_PHASE_IRREGULAR:
// forced end to wireless functionality
            wcmResult = CallFunction(WCM_NOTIFY_TERMINATE);
            break;

        case WCM_PHASE_FATAL_ERROR:
        default:
            FreezeThread();
        }
    }

// ends the WCM library
    wcmResult = WCM_Finish();
    DebugPrintf("WCM_Finish() -> %s\n", gWcmResultText[wcmResult]);
    if (wcmResult != WCM_RESULT_SUCCESS)
    {
        /*
* If an error is returned here, it is highly possible that use of the library is in error.
* It is possible that other threads have already ended the WCM library.
         */
        OS_Free(wcmBuf);
        return;
    }

// release the blocked off memory
    OS_Free(wcmBuf);
    return;
}

/*---------------------------------------------------------------------------*
Name: FreezeThread

Description: Called from the thread, to enter an infinite loop and freeze.

Arguments: None.

Returns: There are no returns from this function.
 *---------------------------------------------------------------------------*/
static void FreezeThread(void)
{
    /*
* As there is nothing that can be done for a FATAL ERROR, set to freeze internally.
* Possible causes include fatal memory corruption, an overloading of ARM7
* causing processes to fail or a thread stack overflow.
* The end user must be shown an appropriate error message, but as
* there is no way in software to recover, the use can only restart the device.
     */
    while (TRUE)
    {
        OS_Sleep(10000);
    }
}

/*---------------------------------------------------------------------------*
Name: CallFunction

Description: Calls the WCM asynchronous API and waits until the asynchronous process completes.
Assumed to be called by a thread, be aware that it will wait it
the completion of the asynchronous process is blocked internally.

Arguments: The WCM asynchronous API types are specifies in the WCM_NOTIFY_* format.
The following are the specifiable parameters:
WCM_NOTIFY_STARTUP      : Runs WCM_StartupAsync.
WCM_NOTIFY_CLEANUP      : Runs WCM_CleanupAsync.
WCM_NOTIFY_BEGIN_SEARCH : Runs WCM_SearchAsycn (start).
WCM_NOTIFY_END_SEARCH   : Runs WCM_SearchAsync (stop).
WCM_NOTIFY_CONNECT      : Runs WCM_ConnectAsync.
WCM_NOTIFY_DISCONNECT   : Runs WCM_DisconnectAsync.
WCM_NOTIFY_TERMINATE    : Runs WCM_TerminateAsync.

Returns: Returns synchronous and asynchronous process results.
The returned value is one of the following:
WCM_RESULT_SUCCESS      : Process completed normally.
WCM_RESULT_FAILURE      : Process failed midway for some reason.
WCM_RESULT_FATAL_ERROR  : beyond assumed scope of parameters
 *---------------------------------------------------------------------------*/
static s32 CallFunction(s16 notifyId)
{
    s32         wcmResult = WCM_RESULT_FATAL_ERROR;
    OSIntrMode  e = OS_DisableInterrupts();

    while (TRUE)
    {
// delineates the API to be run
        switch (notifyId)
        {
        case WCM_NOTIFY_STARTUP:
            wcmResult = WCM_StartupAsync(&gWcmConfig, ControlCallback);
            DebugPrintf("WCM_StartupAsync() -> %s\n", gWcmResultText[wcmResult]);
            break;

        case WCM_NOTIFY_CLEANUP:
            wcmResult = WCM_CleanupAsync();
            DebugPrintf("WCM_CleanupAsync() -> %s\n", gWcmResultText[wcmResult]);
            break;

        case WCM_NOTIFY_BEGIN_SEARCH:
            /****************************************************************/

            /*  place where critical settings are performed for searching for an access point    */
            wcmResult = WCM_BeginSearchAsync((void*)constApInfoPtr->bssId, (void*)constApInfoPtr->essId, (u32) (
                                                 WCM_OPTION_CHANNEL_ALL | gWcmScanType |
                                                 (gWcmManualConnect ? WCM_OPTION_ROUNDSCAN_NOTIFY : WCM_OPTION_ROUNDSCAN_IGNORE)
                                             ));

            /****************************************************************/
            DebugPrintf("WCM_BeginSearchAsync -> %s\n", gWcmResultText[wcmResult]);
            break;

        case WCM_NOTIFY_END_SEARCH:
            wcmResult = WCM_EndSearchAsync();
            DebugPrintf("WCM_EndSearchAsync -> %s\n", gWcmResultText[wcmResult]);
            break;

        case WCM_NOTIFY_CONNECT:
            /****************************************************************/

            /*  place where critical settings are performed for connecting to an access point    */
            wcmResult = WCM_ConnectAsync(&gWcmBssDesc, (void*)(&(constApInfoPtr->wepDesc)),
                                         gWcmPowerMode | constApInfoPtr->auth_option);

            /****************************************************************/
            DebugPrintf("WCM_ConnectAsync() -> %s\n", gWcmResultText[wcmResult]);
            break;

        case WCM_NOTIFY_DISCONNECT:
            wcmResult = WCM_DisconnectAsync();
            DebugPrintf("WCM_DisconnectAsync() -> %s\n", gWcmResultText[wcmResult]);
            break;

        case WCM_NOTIFY_TERMINATE:
            wcmResult = WCM_TerminateAsync();
            DebugPrintf("WCM_TerminateAsync() -> %s\n", gWcmResultText[wcmResult]);
            break;

        default:
            (void)OS_RestoreInterrupts(e);
            return wcmResult;
        }

// confirms synchronous process results
        switch (wcmResult)
        {
        case WCM_RESULT_REJECT:
// retries after waiting a little
            OS_Sleep(1);
            continue;

        case WCM_RESULT_PROGRESS:
        case WCM_RESULT_ACCEPT:
// waits until asynchronous process response is called back
            OS_SleepThread(&gWcmThreadQ);
            break;

        case WCM_RESULT_SUCCESS:
        case WCM_RESULT_FAILURE:
// process is completed synchronously, and asynchronous processes are not performed
            (void)OS_RestoreInterrupts(e);
            return wcmResult;

        case WCM_RESULT_WMDISABLE:
            /*
* Wm library not operating on ARM7.
* You need to launch the wireless driver on ARM7 using the WVR driver before
* beginning to use WCM, but that needs to be performed within the application.
             */
            OS_Panic("ARM7 WM library is not ready.\n");

        case WCM_RESULT_NOT_ENOUGH_MEM: // this returned value will never be returned for an asynchronous function
        case WCM_RESULT_FATAL_ERROR:
        default:
            (void)OS_RestoreInterrupts(e);
            FreezeThread();
        }

// confirms asynchronous process results
        switch (gWcmResult)
        {
        case WCM_RESULT_SUCCESS:
        case WCM_RESULT_FAILURE:
            wcmResult = gWcmResult;
            break;

        case WCM_RESULT_PROGRESS:       // these process results are never passed to asynchronous process results
        case WCM_RESULT_ACCEPT:         // these process results are never passed to asynchronous process results
        case WCM_RESULT_REJECT:         // these process results are never passed to asynchronous process results
        case WCM_RESULT_WMDISABLE:      // these process results are never passed to asynchronous process results
        case WCM_RESULT_NOT_ENOUGH_MEM: // these process results are never passed to asynchronous process results
        case WCM_RESULT_FATAL_ERROR:
        default:
            (void)OS_RestoreInterrupts(e);
            FreezeThread();
        }

        break;
    }

    (void)OS_RestoreInterrupts(e);
    return wcmResult;
}

/*---------------------------------------------------------------------------*
Name: ControlCallback

Description: Callback function to receive WCM asynchronous API process result response and
irregular notifications from the WCM library.

Arguments: arg     -   Pointer to the structure indicating the notification and/or process result.

Returns: None.
 *---------------------------------------------------------------------------*/
static void ControlCallback(WCMNotice* arg)
{
    static s16  lastNotify = -1;
    static s16  lastResult = -1;

//display if result different from before
    if (lastNotify != arg->notify || lastResult != arg->result)
    {
        DebugPrintf(" cb %s %s\n", gWcmNotifyText[arg->notify], gWcmResultText[arg->result]);
        lastNotify = arg->notify;
        lastResult = arg->result;
    }

    switch (arg->notify)
    {
    case WCM_NOTIFY_STARTUP:
        gWcmSearchAroundCount = 0;

    case WCM_NOTIFY_CLEANUP:
    case WCM_NOTIFY_BEGIN_SEARCH:
    case WCM_NOTIFY_END_SEARCH:
    case WCM_NOTIFY_CONNECT:
    case WCM_NOTIFY_TERMINATE:
// asynchronous process result response in a 1:1 correspondence with requests
        gWcmResult = arg->result;
        OS_WakeupThread(&gWcmThreadQ);
        break;

    case WCM_NOTIFY_DISCONNECT:
#ifdef SDK_THREAD_INFINITY
        if (gWcmThreadQ.head != NULL)
#else
            if (gWcmThreadQ != 0)
#endif
            {
// asynchronous process result response in a 1:1 correspondence with requests
                gWcmResult = arg->result;
                OS_WakeupThread(&gWcmThreadQ);
            }
            else
            {
// when a stop state occurs while a thread is running or when waiting for a message
                if (arg->result == WCM_RESULT_SUCCESS)
                {
// notification indicating AP disconnect during DCF communication
                    (void)OS_SendMessage(&gWcmMessageQ, LOCAL_WCM_MSG_DISCONNECTED, OS_MESSAGE_NOBLOCK);
                }
                else
                {
// notification indicating some sort of error occurrence during DCF communication
                    (void)OS_JamMessage(&gWcmMessageQ, LOCAL_WCM_MSG_ERROR, OS_MESSAGE_NOBLOCK);
                }
            }
        break;

    case WCM_NOTIFY_FOUND_AP:
// asynchronous notification during auto-search
        {
            if (arg->result == WCM_RESULT_SUCCESS)
            {
                OSMessage   tempMessage;

// notification indicating found AP during auto-search
                if (OS_ReadMessage(&gWcmMessageQ, &tempMessage, OS_MESSAGE_NOBLOCK) == FALSE)
                {
// as notifications would take place multiple times, limited to only occurring when the message queue is empty
                    (void)OS_SendMessage(&gWcmMessageQ, LOCAL_WCM_MSG_FOUND_AP, OS_MESSAGE_NOBLOCK);
                }
            }
            else
            {
// notification indicating some sort of error occurrence during auto-search
                (void)OS_JamMessage(&gWcmMessageQ, LOCAL_WCM_MSG_ERROR, OS_MESSAGE_NOBLOCK);
            }
        }
        break;

    case WCM_NOTIFY_SEARCH_AROUND:
// irregular notification indication a full loop on the specified channel during auto-search
        ++gWcmSearchAroundCount;
        break;
    }
}

/*---------------------------------------------------------------------------*
Name: InitWcmApInfo

Description: Gets access point connection parameters from environment setting values
and initializes.

Arguments: apinfo   connection parameter
apclass  access point class name

Returns: TRUE   success
FALSE  failure
 *---------------------------------------------------------------------------*/
static BOOL myEnvCopy(const char* name, void* dest, int dest_size);
static BOOL myEnvCopy(const char* name, void* dest, int dest_size)
{
    void*   src;
    int     src_size;

    if (ENV_GetBinaryAndSize(name, &src, &src_size))
    {
        MI_CpuCopy8(src, dest, (u32) MATH_MIN(src_size, dest_size));
        return TRUE;
    }

    return FALSE;
}

BOOL InitWcmApInfo(WcmControlApInfo* apinfo, const char* apclass)
{
    BOOL    isvalid;

    ENV_SetClass(apclass);

    if (ENV_GetBOOL(".ISVALID", &isvalid) && isvalid)
    {
        MI_CpuClear8(apinfo, sizeof(WcmControlApInfo));

        if (myEnvCopy(".ESSID", apinfo->essId, sizeof(apinfo->essId)) &&
            myEnvCopy(".BSSID", apinfo->bssId, sizeof(apinfo->bssId)) && ENV_GetU32(".AUTHMODE", &apinfo->auth_option) &&
            ENV_GetU8(".WEP.MODE", &apinfo->wepDesc.mode) && ENV_GetU8(".WEP.KEYID", &apinfo->wepDesc.keyId) &&
            myEnvCopy(".WEP.KEY", apinfo->wepDesc.key, sizeof(apinfo->wepDesc.key)))
        {
            return TRUE;
        }
    }

    return FALSE;
}

/*---------------------------------------------------------------------------*
End of file
 *---------------------------------------------------------------------------*/
