/*---------------------------------------------------------------------------*
Project: NitroWiFi - demo - soc - shared
File: ftp.c

Copyright 2006 Nintendo.  All rights reserved.

These coded instructions, statements, and computer programs contain
proprietary information of Nintendo of America Inc. and/or Nintendo
Company Ltd., and are protected by Federal copyright law.  They may
not be disclosed to third parties or copied or duplicated in any form,
in whole or in part, without the prior written consent of Nintendo.

$Log: ftp.c,v $
Revision 1.2  03/10/2006 09:22:43  kitase_hirotake
  INDENT SOURCE

Revision 1.1  01/12/2006 02:16:41  okubata_ryoma
Added ftp.c


$NoKeywords: $
 *---------------------------------------------------------------------------*/
#include <nitro.h>
#include "ftp.h"

static int      FTPi_CommandSocket = -1;
static int      FTPi_DataSocket = -1;

static char     send_buf[FTPi_BUFFERSIZE_UP];

int         FTPi_Readline(char* buffer, size_t size);
int         FTPi_Printf(const char* fmt, ...);
int         FTPi_GetStatus(const char* line);
int         ConnectToFtpServer(const char* http_addr, u16 http_port);
int         FTP_TransferType(const char type);
BOOL        FTPi_ReadMoreData(char* line);
BOOL        FTPi_IsMoreDataAvailable(const char* line);
BOOL        FTPi_IsSuccess(const char* line);

static char*    Upload_fgets(char* buf, int len, FSFile* p_file);

/*****************************************************************************/

/* function */

/*---------------------------------------------------------------------------*
Name: FTP_Open

Description: Connects to the FTP server.

Arguments: http_addr     : Connection target address
http_port     : Connection target port

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_Open(const char* http_addr, u16 http_port)
{
// Makes a connection to the server for commands.
    FTPi_CommandSocket = ConnectToFtpServer(http_addr, http_port);

    return FTPi_CommandSocket;
}

/*---------------------------------------------------------------------------*
Name: FTP_Login

Description: Issues the USER and PASS commands and logs-in to the FTP server.

Arguments: user_name     : Connection target account
password: Connection target password
                

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_Login(const char* user_name, const char* password)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];

// Receives the message for Connect.
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

// UserName
    FTPi_RETURN_ERROR(FTPi_Printf("USER %s\r\n", user_name));
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

// Password
    FTPi_RETURN_ERROR(FTPi_Printf("PASS %s\r\n", password));
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTP_Passive

Description: Establishes a data connection in Passive mode.

Arguments: filetype File transmission mode TRUE  : BINARY mode
FALSE : ASCII mode

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_Passive(BOOL filetype)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];

// FileType
    if (filetype)
    {
        FTPi_RETURN_ERROR(FTP_TransferType('I'));
    }
    else
    {
        FTPi_RETURN_ERROR(FTP_TransferType('A'));
    }

// Passive mode
    FTPi_RETURN_ERROR(FTPi_Printf("PASV\r\n"));
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

// Extracts the server address and port from the message corresponding to the PASV command.
    {
        static char http_addr[16];
        static u16  http_port;
        char*       tmp_ptr;

        tmp_ptr = strchr(recv_buf, '(');
        if (tmp_ptr == NULL)
        {
            OS_TPrintf("recv_buf = %s\n", recv_buf);
            return -1;
        }

        tmp_ptr++;
        (void)strncpy(http_addr, tmp_ptr, 1);
        http_addr[1] = '\0';
        {
            int count = 0;  // The number of times ',' appears.
            while (1)
            {
                tmp_ptr++;
                if (tmp_ptr[0] == ',')
                {
                    count++;
                    if (count == 4)
                    {
                        break;
                    }
                    (void)strncat(http_addr, ".", 1);
                }
                else
                {
                    (void)strncat(http_addr, &tmp_ptr[0], 1);
                }
            }

            tmp_ptr++;
            http_port = (u16) atoi(tmp_ptr);
            while (tmp_ptr[0] != ',')
            {
                tmp_ptr++;
            }

            tmp_ptr++;
            http_port = (u16) (http_port * 256 + atoi(tmp_ptr));
        }

// Makes a connection for data to the server
        FTPi_DataSocket = ConnectToFtpServer(http_addr, http_port);
    }

    return FTPi_DataSocket;
}

/*---------------------------------------------------------------------------*
Name: FTP_OnlyStor

Description: Gets the file size for upload and sends the STOR command.

Arguments: filename: Filename for upload.

Returns: The file size for upload: if >= 0, returns TRUE.
 *---------------------------------------------------------------------------*/
int FTP_OnlyStor(const char* filename)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];
    int     buf_len;    // Length of a string in the file.
    char*   send_image;
    u32     open_filesize = 0;
    char*   return_point;

    FSFile  file;
    BOOL    open_is_ok;
    FS_InitFile(&file);

    send_image = send_buf;

    open_is_ok = FS_OpenFile(&file, filename);
    OS_TPrintf("FS_OpenFile(\"%s\") ... %s!\n", filename, open_is_ok ? "OK" : "ERROR");
    open_filesize = FS_GetLength(&file);
    if (open_is_ok)
    {
        char    buf[FILEDATA256];
        MI_CpuClear8(buf, sizeof(buf));
        MI_CpuClear8(send_buf, sizeof(send_buf));
        OS_TPrintf("\n--------------------------------\n");
        while (Upload_fgets(buf, sizeof(buf), &file) && (*buf != '\0'))
        {
            MI_CpuCopy8(buf, send_buf, sizeof(buf));
            buf_len = STD_StrLen(buf);
            send_image += buf_len;
            return_point = buf;
            while ((return_point = STD_StrStr(buf, "\r")) != NULL)
            {
                *return_point = ' ';
            }

            OS_TPrintf("  %s", buf);
        }

        OS_TPrintf("\n--------------------------------\n");
        send_image -= buf_len;
    }

// Sends the RETR command.
    FTPi_RETURN_ERROR(FTPi_Printf("STOR %s\r\n", filename));
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

    return (int)open_filesize;
}

/*---------------------------------------------------------------------------*
Name: FTP_OnlySendData

Description: Asynchronously runs SOC_Send once.

Arguments: size   : Size to send.

Returns: Size uploaded with this send.
 *---------------------------------------------------------------------------*/
int FTP_OnlySendData(int size)
{
    int result;

    size = MATH_IMin(size, FTPi_BUFFERSIZE_UP);

    result = SOC_Send(FTPi_DataSocket, send_buf, size, SOC_MSG_DONTWAIT);

// SOC_EAGAIN and SOC_ENOBUFS return a FALSE value, but since they are not errors, they are omitted.
    if ((result != SOC_EAGAIN) && (result != SOC_ENOBUFS))
    {
        FTPi_RETURN_ERROR(result);
        return (int)result;
    }

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTP_OnlyRetr

Description: Only sends the RETR command

Arguments: filename: Filename for download.

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_OnlyRetr(const char* filename)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];

// Sends the RETR command.
    FTPi_RETURN_ERROR(FTPi_Printf("RETR %s\r\n", filename));
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTP_OnlyRecvData

Description: Asynchronously runs SOC_Recv once.

Arguments: filesize    : Data transmission size.
recv_filename    : When using the NLST command, enter the received file name.

Returns: Size downloaded with this recv.
 *---------------------------------------------------------------------------*/
int FTP_OnlyRecvData(int filesize, char recv_filename[FILEDATA256])
{
    char    recv_buf[FILEDATA256];
    char    print_buf[FILEDATA256];
    char*   return_point;
    int     result;
    char*   precv_buf;
    int     filename_len;

    MI_CpuClear8(recv_buf, FILEDATA256);    // Initializes the buffer.
    result = SOC_Recv(FTPi_DataSocket, recv_buf, filesize, SOC_MSG_DONTWAIT);

    if (recv_buf[0] != NULL)
    {
        if (recv_filename)
        {
            MI_CpuCopy8(recv_buf, recv_filename, FILEDATA256);
        }
        else
        {
            precv_buf = recv_buf;
            while (*precv_buf)
            {
                return_point = STD_StrStr(precv_buf, "\n");
                *(return_point - 1) = ' ';
                filename_len = (int)(return_point - precv_buf);
                (void)STD_StrLCpy(print_buf, precv_buf, filename_len + 1);

                OS_TPrintf("  %s\n", print_buf);
                precv_buf += filename_len + 1;
            }
        }
    }

// SOC_EAGAIN and SOC_ENOBUFS return a FALSE value, but since they are not errors, they are omitted.
    if ((result != SOC_EAGAIN) && (result != SOC_ENOBUFS))
    {
        FTPi_RETURN_ERROR(result);
        return (int)result;
    }

    return result;
}

/*---------------------------------------------------------------------------*
Name: FTP_CloseDataSocket

Description: Closes FTPi_DataSocket and disconnects.

Arguments: none.

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_CloseDataSocket(void)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];
    int     result;

// Closes the socket for data and disconnects.
    result = SOC_Close(FTPi_DataSocket);

    FTPi_DataSocket = -1;
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTP_OnlyAbor

Description: Only sends the ABOR command (doesn't even receive a corresponding message).

Arguments: none.

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_OnlyAbor()
{
// Sends the RETR command.
    FTPi_RETURN_ERROR(FTPi_Printf("ABOR\r\n"));

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTP_OnlySendCommand

Description: Performs FTPi_Print.

Arguments: fmt   : String to write.


Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_OnlySendCommand(const char* fmt, ...)
{
// Sends the RETR command.
    FTPi_RETURN_ERROR(FTPi_Printf(fmt));

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTP_OnlyRecvCommand

Description: Runs FTPi_Readline

Arguments: none.

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_OnlyRecvCommand(void)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];

    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTP_FileSize

Description: Gets the size of the specified file name.

Arguments: filename: File name to examine.

Returns: Size of the file.
 *---------------------------------------------------------------------------*/
int FTP_FileSize(const char* filename)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];
    char*   tmp_ptr;

// Checks the size of the file to download.
    FTPi_RETURN_ERROR(FTPi_Printf("SIZE %s\r\n", filename));
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

// Gets the size of the files to DownLoad from the SIZE command's acknowledgement message.
    tmp_ptr = strchr(recv_buf, ' ');
    if (tmp_ptr == NULL)
    {
        return -1;
    }

    tmp_ptr++;

    return atoi(tmp_ptr);
}

/*---------------------------------------------------------------------------*
Name: FTP_Quit

Description: Quits the FTP connection.

Arguments: None

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_Quit(void)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];
    int     result;

    FTPi_RETURN_ERROR(FTPi_Printf("QUIT\r\n"));
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

    result = SOC_Close(FTPi_CommandSocket);

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTPi_Readline

Description: Reads one line from the command connection up to a line break,
                then copies it to the buffer.

Arguments: buffer  : A pointer to the buffer.
size   : Size of the buffer.


Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTPi_Readline(char* buffer, size_t size)
{
    int result;
    int i;
    int buffer_size = 0;

    for (i = 0; i < size; i++)
    {
        result = SOC_Read(FTPi_CommandSocket, buffer + i, 1);
        if ((result == 0) || (buffer[i] == '\n'))
        {
            buffer[i + 1] = '\0';
            break;
        }

        buffer_size++;
    }

// Displays the receive command.
    buffer = STD_StrLCpy(buffer, buffer, buffer_size);
    OS_TPrintf("<%s\n", buffer);
    return i;
}

/*---------------------------------------------------------------------------*
Name: FTPi_Printf

Description: Writes to command connection.

Arguments: fmt   : String to write.


Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTPi_Printf(const char* fmt, ...)
{
    char    buffer[FTPi_BUFFERSIZE_COMMAND];
    {
        va_list vlist;

        va_start(vlist, fmt);
        (void)OS_VSNPrintf(buffer, FTPi_BUFFERSIZE_COMMAND, fmt, vlist);    // Copies to buffer
        va_end(vlist);
    }
    {
        int result;

//write operation
        result = SOC_Write(FTPi_CommandSocket, buffer, (int)strlen(buffer));

// Displays the send command.
        (void)STD_StrLCpy(buffer, buffer, (int)strlen(buffer) - 1);
        OS_TPrintf(">%s\n", buffer);
        return result;
    }
}

/*---------------------------------------------------------------------------*
Name: FTPi_GetStatus

Description: Gets the status code from the server's acknowledgement string.

Arguments: line : The server's acknowledgement string.


Returns: Status code.
 *---------------------------------------------------------------------------*/
int FTPi_GetStatus(const char* line)
{
    char    str[5];
    int     recv_command;

// Gets the acknowledgement code.
    (void)STD_StrLCpy(str, line, 4);    //STD_
    recv_command = atoi(str);

    return recv_command;
}

/*---------------------------------------------------------------------------*
Name: ConnectToFtpServer

Description: Creates a socket, tries to connect to the FTP server, and waits for completion.

Arguments: http_addr     : Connection target address
http_port     : Connection target port

Returns: The created and connected socket ( >= 0) or an error value ( < 0).
 *---------------------------------------------------------------------------*/
int ConnectToFtpServer(const char* http_addr, u16 http_port)
{
    int         socket;
    int         result;

    SOHostEnt*  host_ent;

    /* Creates a socket. */
    socket = -1;
    result = SOC_Socket(SOC_PF_INET, SOC_SOCK_STREAM, 0);
    if (result < 0)
        OS_TPrintf("failed to create socket\n");
    {
        SOSockAddrIn    sa_in;

        socket = result;

        /* Loads host information. */
        sa_in.len = sizeof(sa_in);
        sa_in.family = IP_INET;
        sa_in.port = SOC_HtoNs(http_port);
        host_ent = SOC_GetHostByName(http_addr);
        MI_CpuCopy8(host_ent->addrList[0], &sa_in.addr, IP_ALEN);

        /* Tries to connect. */
        result = SOC_Connect(socket, &sa_in);
        if (result < 0)
            OS_TPrintf("failed to connect\n");
    }

    if ((result < 0) && (socket >= 0))
    {
        result = SOC_Close(socket);
        if (result < 0)
            OS_TPrintf("failed to close socket\n");
    }

    return(result >= 0) ? socket : result;
}

/*---------------------------------------------------------------------------*
Name: FTP_TransferType

Description: Sets the data display method for FTP to binary.

Arguments: type: Specifies TYPE.
'I' : Binary.
'A' : ASCII

Returns: Values >= 0 indicate success.
 *---------------------------------------------------------------------------*/
int FTP_TransferType(const char type)
{
    char    recv_buf[FTPi_BUFFERSIZE_COMMAND];

// Selects FileType.
    FTPi_RETURN_ERROR(FTPi_Printf("TYPE %c\r\n", type));
    FTPi_RETURN_ERROR(FTPi_Readline(recv_buf, FTPi_BUFFERSIZE_COMMAND));

// Checks the ACK command.
    if (!FTPi_IsSuccess(recv_buf))
    {
        return -1;
    }

// Continues to read to the end of multi-line acknowledgements.
    if (!FTPi_ReadMoreData(recv_buf))
    {
        return -1;
    }

    return 0;
}

/*---------------------------------------------------------------------------*
Name: FTPi_ReadMoreData

Description: Continues to read until the end when multi-line acknowledgements are returned from the FTP server.

Arguments: line : Load buffer.

Returns: Successful if TRUE.
 *---------------------------------------------------------------------------*/
BOOL FTPi_ReadMoreData(char* line)
{
    int result;

// Continues to read to the end of multi-line acknowledgements.
    while (FTPi_IsMoreDataAvailable(line))
    {
        result = FTPi_Readline(line, FTPi_BUFFERSIZE_COMMAND);
        if (result <= 0)
        {
            return FALSE;
        }

// Checks the ACK command.
        if (!FTPi_IsSuccess(line))
        {
            return FALSE;
        }
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*
Name: FTPi_IsMoreDataAvailable

Description: Checks whether there is a multi-line acknowledgement from the FTP server.

Arguments: line : Load buffer.

Returns: Successful if TRUE.
 *---------------------------------------------------------------------------*/
BOOL FTPi_IsMoreDataAvailable(const char* line)
{
    if (line[3] == '-')
    {
        return TRUE;
    }

    return FALSE;
}

/*---------------------------------------------------------------------------*
Name: FTPi_IsSuccess

Description: Checks whether the ACK command from the FTP server is valid.

Arguments: line : The buffer, for which Read was executed, that contains the ACK command.

Returns: Successful if TRUE.
 *---------------------------------------------------------------------------*/
BOOL FTPi_IsSuccess(const char* line)
{
    int result;

// Checks the ACK command.
    result = FTPi_GetStatus(line);
    if (result >= 400)
    {
        return FALSE;
    }

    return TRUE;
}

/*---------------------------------------------------------------------------*
Name: FTP_Upload

Description: Performs file upload.

Arguments: pfiletype    : File transmission mode.
pfilesize    : Data transmission size.
pfilename    : File name

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_Upload(void* pfiletype, void* pfilesize, void* pfilename)
{
    BOOL        filetype = (BOOL) pfiletype;
    int         filesize = (int)pfilesize;
    const char*     filename = (const char*)pfilename;
    char        cur_key;                        // For the FTP command's send request detection button.
    int     result;
    int     passive_result;

    OSTick  timerStart, timerEnd;               // Tic to measure throughput.
    u64     transTime;                          // Time taken for send/receive.
    int     transSize;                          // Size sent/received.
    BOOL    abor_flg = FALSE;                   // Update suspend flag.
    int     sendSize = FTPi_BUFFERSIZE_DATA;    // Size of data being sent
    int     restSize;               // Remaining size to be sent.

// The data connection's socket is returned to this result.
    passive_result = FTP_Passive(filetype);
    if (passive_result < 0)
        OS_TPrintf("failed in FTP_Passive\n");

// Displays file data.
    filesize = FTP_OnlyStor(filename);
    if (filesize < 0)
        OS_TPrintf("failed in FTP_OnlyStor\n");

    restSize = filesize;

// Gets the time and calculates receive time.
    transTime = 0;                  // Initializes communications time.
    transSize = 0;                  // Initializes total send size.
    timerStart = OS_GetTick();

    if (passive_result >= 0 && filesize >= 0)
    {
        do
        {
            sendSize = MATH_IMin(restSize, FTPi_BUFFERSIZE_DATA);

            result = FTP_OnlySendData(sendSize);
            transSize += result;    // Size of sent files.
            if (filesize <= transSize)
            {
                break;
            }
            else
            {
                cur_key = (char)PAD_Read();
                if (cur_key == PAD_BUTTON_B)
                {
// The B button was pressed, so the receive operation is terminated before completion.
                    (void)FTP_OnlyAbor();
                    abor_flg = TRUE;
                    break;
                }
            }

            restSize -= result;
        }
        while (1);

        timerEnd = OS_GetTick();
        transTime = OS_TicksToMilliSeconds((timerEnd - timerStart));

        OS_TPrintf("\nsendTime = %d msec\n", transTime);
        OS_TPrintf("this phase send result = %d byte\n", transSize);
        OS_TPrintf("throughput = %d byte/sec\n\n", transSize * 1000 / transTime);

// Closes the data connection.
        result = FTP_CloseDataSocket();
        if (result < 0)
            OS_TPrintf("failed in FTP_CloseDataSocket\n");
    }

    if (abor_flg == TRUE)
    {
// Receive for ABOR.
        result = FTP_OnlyRecvCommand();
    }
}

/*---------------------------------------------------------------------------*
Name: FTP_Download

Description: Performs file download.

Arguments: pfiletype    : File transmission mode.
pfilesize    : Data transmission size.
pfilename    : File name

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_Download(void* pfiletype, void* pfilesize, void* pfilename)
{
    BOOL        filetype = (BOOL) pfiletype;
    int         filesize = (int)pfilesize;
    const char*     filename = (const char*)pfilename;
    char        cur_key;            // For the FTP command's send request detection button.
    int     result;
    int     passive_result;
    int     retr_result;

    OSTick  timerStart, timerEnd;   // Tic to measure throughput.
    u64     transTime;              // Time taken for send/receive.
    int     transSize;              // Size sent/received.
    BOOL    abor_flg = FALSE;       // Update suspend flag.
    int     recvSize = FTPi_BUFFERSIZE_DOWN;    // Receive size.
    int     restSize;               // Size of remaining data to be received.

// The data connection's socket is returned to this result.
    passive_result = FTP_Passive(filetype);
    if (passive_result < 0)
        OS_TPrintf("failed in FTP_Passive\n");

    retr_result = FTP_OnlyRetr(filename);
    if (retr_result < 0)
        OS_TPrintf("failed in FTP_OnlyRetr\n");
    restSize = filesize;

// Gets the time and calculates receive time.
    transTime = 0;                  // Initializes communications time.
    transSize = 0;                  // Initializes receive size.
    timerStart = OS_GetTick();

    if (passive_result >= 0 && retr_result >= 0)
    {
        OS_TPrintf("\n--------------------------------\n");
        do
        {
            recvSize = MATH_IMin(restSize, FTPi_BUFFERSIZE_DOWN);
            result = FTP_OnlyRecvData(recvSize, NULL);

// SOC_EAGAIN and SOC_ENOBUFS return a FALSE value, but since they are not errors, they are omitted.
            if ((result == SOC_EAGAIN) || (result == SOC_ENOBUFS))
                result = 0;
            if (result < 0)
                OS_TPrintf("failed in FTP_OnlyRecvData\n");
            transSize += result;    // Size of received files.
            if (recvSize <= transSize)
            {
                break;
            }
            else
            {
                cur_key = (char)PAD_Read();
                if (cur_key == PAD_BUTTON_B)
                {
// The B button was pressed, so the receive operation is terminated before completion.
                    (void)FTP_OnlyAbor();
                    abor_flg = TRUE;
                    break;
                }
            }

            restSize -= result;
        }
        while (1);
        OS_TPrintf("\n--------------------------------\n");

        timerEnd = OS_GetTick();
        transTime = OS_TicksToMilliSeconds((timerEnd - timerStart));

        OS_TPrintf("\nrecvTime = %d msec\n", transTime);
        OS_TPrintf("this phase recv result = %d\n", transSize);
        OS_TPrintf("throughput = %d byte/sec\n\n", transSize * 1000 / transTime);

// Closes the data connection.
        result = FTP_CloseDataSocket();
        if (result < 0)
            OS_TPrintf("failed in FTP_CloseDataSocket\n");
    }

    if (abor_flg == TRUE)
    {
// Receives ACK commands for ABOR.
        result = FTP_OnlyRecvCommand();
    }
}

/*---------------------------------------------------------------------------*
Name: FTP_Nlst

Description: Displays a list of files.

Arguments: pfiletype    : File transmission mode.
pfilesize    : Data transmission size (unused).
pfilename    : Filename (unused).

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_Nlst(void* pfiletype, void* pfilesize, void* pfilename)
{
    BOOL    filetype = (BOOL) pfiletype;

    int     result;
    int     passive_result;
    int     send_result;

    (void)pfilesize;
    (void)pfilename;

// The data connection's socket is returned to this result.
    passive_result = FTP_Passive(filetype);
    if (passive_result < 0)
        OS_TPrintf("failed in FTP_Passive\n");

    send_result = FTP_OnlySendCommand("NLST -1\r\n");
    if (send_result < 0)
        OS_TPrintf("failed in NLST\n");

    if (passive_result >= 0 && send_result >= 0)
    {
        OS_TPrintf("\n--------------------------------\n");
        while (FTP_OnlyRecvData(FILEDATA256, NULL) != 0)
        {
// Loads all data.
            ;
        }

        OS_TPrintf("\n--------------------------------\n");

        result = FTP_OnlyRecvCommand();
        if (result < 0)
            OS_TPrintf("failed in Command recv\n");

        result = FTP_CloseDataSocket();
        if (result < 0)
            OS_TPrintf("failed in FTP_CloseDataSocket\n");
    }
}

/*---------------------------------------------------------------------------*
Name: FTP_Logout

Description: Logout

Arguments: pfiletype    : File transmission mode (unused).
pfilesize    : Data transmission size (unused).
pfilename    : Filename (unused).

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_Logout(void* pfiletype, void* pfilesize, void* pfilename)
{
    int result;
    (void)pfiletype;
    (void)pfilesize;
    (void)pfilename;

    result = FTP_Quit();
    if (result < 0)
        OS_TPrintf("failed in FTP_Quit\n");

    OS_Printf("finished ftpclient\n");

    ClearScreen();
}

/*---------------------------------------------------------------------------*
Name: FTP_ASCMode

Description: Changed the file transmission mode to ASCII mode.

Arguments: pfiletype    : File transmission mode.
pfilesize    : Data transmission size (unused).
pfilename    : Filename (unused).

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_ASCMode(void* pfiletype, void* pfilesize, void* pfilename)
{
    BOOL    filetype = (BOOL) pfiletype;

    (void)pfilesize;
    (void)pfilename;

    filetype = FALSE;
    OS_TPrintf("\n--------------------------------\n");
    OS_TPrintf("  file type = ASCII\n");
    OS_TPrintf("\n--------------------------------\n");
}

/*---------------------------------------------------------------------------*
Name: FTP_BINMode

Description: Changed the file transmission mode to BINARY mode.

Arguments: pfiletype    : File transmission mode.
pfilesize    : Data transmission size (unused).
pfilename    : Filename (unused).

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_BINMode(void* pfiletype, void* pfilesize, void* pfilename)
{
    BOOL    filetype = (BOOL) pfiletype;

    (void)pfilesize;
    (void)pfilename;

    filetype = TRUE;
    OS_TPrintf("\n--------------------------------\n");
    OS_TPrintf("  file type = BINARY\n");
    OS_TPrintf("\n--------------------------------\n");
}

/*---------------------------------------------------------------------------*
Name: FTP_Ls

Description: Displays a list of files.

Arguments: pfiletype    : File transmission mode (unused).
pfilesize    : Data transmission size (unused).
pfilename    : Filename (unused).

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_Ls(void* pfiletype, void* pfilesize, void* pfilename)
{
    FSFile      d;
    FSDirEntry  e;

    (void)pfiletype;
    (void)pfilesize;
    (void)pfilename;

    FS_InitFile(&d);
    (void)FS_FindDir(&d, "");

    OS_TPrintf("\n--------------------------------\n");
    MI_WriteByte(e.name + 0, (u8) '\0');
    (void)FS_TellDir(&d, &e.dir_id);
    (void)RecursiveEnumDirEx(0, &e, TRUE, NULL);
    OS_TPrintf("\n--------------------------------\n");
}

/*---------------------------------------------------------------------------*
Name: FTP_Help

Description: Displays help.

Arguments: pfiletype    : File transmission mode (unused).
pfilesize    : Data transmission size (unused).
pfilename    : Filename (unused).

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_Help(void* pfiletype, void* pfilesize, void* pfilename)
{
    (void)pfiletype;
    (void)pfilesize;
    (void)pfilename;

    OS_TPrintf("\n--------------------------------\n");
    OS_TPrintf("  Push UP       : Cursor Up\n");
    OS_TPrintf("  Push DOWN     : Cursor Down\n");
    OS_TPrintf("  Push A Button : Menu Select\n");
    OS_TPrintf("  Push B Button : Cancel\n");
    OS_TPrintf("\n--------------------------------\n");
}

/*---------------------------------------------------------------------------*
Name: FTP_LIST

Description: Displays a list of files.

Arguments: pfiletype    : File transmission mode.
pfilename    : Filenames used only with the LIST command.
download_filedata    : Data from the file to be downloaded.
option               FALSE : NLST -1
TRUE  : LIST filename

Returns: None.
 *---------------------------------------------------------------------------*/
void FTP_LIST(void* pfiletype, void* pfilename, char download_filedata[FILEDATA256], int option)
{
    BOOL        filetype = (BOOL) pfiletype;

    const char*     filename;
    char        recv_filedata[FILEDATA256];

    int result;
    int passive_result;
    int send_result;

    MI_CpuClear8(recv_filedata, sizeof(recv_filedata));

    if (option)
    {
        filename = (const char*)pfilename;
    }
    else
    {
        (void)pfilename;
    }

// The data connection's socket is returned to this result.
    passive_result = FTP_Passive(filetype);
    if (passive_result < 0)
        OS_TPrintf("failed in FTP_Passive\n");

// When option is TRUE: LIST command
    if (option)
    {
        send_result = FTP_OnlySendCommand("LIST %s\r\n", filename);
    }

// When option is FALSE: NLST -1 command
    else
    {
        send_result = FTP_OnlySendCommand("NLST -1\r\n");
    }

    if (send_result < 0)
        OS_TPrintf("failed in NLST\n");

    if (passive_result >= 0 && send_result >= 0)
    {
        while (FTP_OnlyRecvData(FILEDATA256, recv_filedata) != 0)
        {
// Loads all data.
            if (recv_filedata[0] != NULL)
            {
                (void)STD_StrCat(download_filedata, recv_filedata);
            }

            MI_CpuClear8(recv_filedata, sizeof(recv_filedata));
        }

        result = FTP_OnlyRecvCommand();
        if (result < 0)
            OS_TPrintf("failed in Command recv\n");

        result = FTP_CloseDataSocket();
        if (result < 0)
            OS_TPrintf("failed in FTP_CloseDataSocket\n");
    }
}

/*---------------------------------------------------------------------------*
Name: RecursiveEnumDirEx

Description: Displays and saves a list of the filenames in the directory.

Arguments: tab                : Width of the subscript.
pe                 : The address of the target directory and the FSDirEntry structure.
output_main        : If TRUE: Displays the directories and filenames below that directory.
If FALSE: Saves the filenames in that directory.
upload_filename    : Destination of the filename.

Returns: The number of files displayed or saved.
 *---------------------------------------------------------------------------*/
int RecursiveEnumDirEx(int tab, FSDirEntry* pe, BOOL output_main, char upload_filename[FILEDATA256])
{
    int     line = 0;
    FSFile  d;
    FS_InitFile(&d);

    if (output_main)
        OS_Printf("%*s%s/\n", tab, "", pe->name);
    if (FS_SeekDir(&d, &pe->dir_id))
    {
        tab += 2;
        while (FS_ReadDir(&d, pe))
        {
            if (pe->is_directory)
            {
                if (output_main)
                    (void)RecursiveEnumDirEx(tab, pe, output_main, NULL);
            }
            else
            {
                if (output_main)
                    OS_Printf("%*s%s\n", tab, "", pe->name);
                else
                {
                    PrintString(tab, line + (UPLOAD + 1), COLOR_WHITE, "%s", pe->name);

// A line break character is entered as a delimiter.
                    (void)STD_StrCat(pe->name, "\r\n");
                    (void)STD_StrCat(upload_filename, pe->name);
                }
            }

            line++;
        }
    }

    return line;
}

/*---------------------------------------------------------------------------*
Name: Upload_fgets

Description: Gets the file data to upload.

Arguments: buf : Pointer to the destination buffer to which file data is to be saved.
len    : Size of the buffer.
p_file : Address of the FSFile structure.

Returns: File data for upload. Fails if NULL.
 *---------------------------------------------------------------------------*/
static char* Upload_fgets(char* buf, int len, FSFile* p_file)
{
    int ret = FS_ReadFile(p_file, buf, len - 1);
    if (ret < 0)
        return NULL;
    else
    {
        int i;
        for (i = 0; i < ret; ++i)
        {
            u32 c = MI_ReadByte(buf + i);
            if (c == '\n')
            {
                ++i;
                break;
            }
        }

        MI_WriteByte(buf + i, (u8) '\0');
        (void)FS_SeekFile(p_file, i - ret, FS_SEEK_CUR);
    }

    return buf;
}

/*---------------------------------------------------------------------------*/

/* END OF FILE */
