/*---------------------------------------------------------------------------*
Project: NitroWiFi - WCM - demos - wcm-1
File: main.c

Copyright 2005.2006 Nintendo.  All rights reserved.

These coded instructions, statements, and computer programs contain
proprietary information of Nintendo of America Inc. and/or Nintendo
Company Ltd., and are protected by Federal copyright law.  They may
not be disclosed to third parties or copied or duplicated in any form,
in whole or in part, without the prior written consent of Nintendo.

$Log: main.c,v $
Revision 1.8  03/10/2006 09:22:43  kitase_hirotake
  INDENT SOURCE

Revision 1.7  10/25/2005 04:19:19  terui
SVC_WaitVBlankIntr -> OS_WaitVBlankIntr

Revision 1.6  09/05/2005 11:27:43  terui
Changed to perform search using passive scan.

Revision 1.5  07/21/2005 09:32:14  terui
Fix comment.

Revision 1.4  07/21/2005 08:41:08  terui
added debugging display

Revision 1.3  07/18/2005 02:27:13  terui
Fix comment.

Revision 1.2  07/13/2005 00:50:06  terui
Fix comment.

Revision 1.1  07/11/2005 12:07:09  terui
new additions

$NoKeywords: $
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*
Sample that connects to a general wireless access point (AP) using
the WCM library. For a fixed AP with the indicated BSSID, ESSID, 802.11 verification method and
WEP encryption settings indicated herein, connects in DCF mode.
As the AP data is indicated in NitroWiFi/build/demos/shared/ap_info.c,
overwrite the data as needed for your use. If BSSID and ESSID are not fixed, but
optional values, an attempt to connect to the first AP found will be made.

HOWTO: (1) After launching, press a button to connect to the AP.
(2) During connection, press a button to cancel the connection.
(3) Once connection is complete, press a button to disconnect from the AP.
(4) Once connection is complete, if the connection to the AP is broken, the system will automatically attempt to re-connect.
 *---------------------------------------------------------------------------*/
#include <nitro.h>
#include <nitroWiFi.h>
#include "screen.h"
#include "wcm_control.h"

/*---------------------------------------------------------------------------*
constant definitions
 *---------------------------------------------------------------------------*/
#define KEY_REPEAT_START    25  // number of frames until key repeat starts
#define KEY_REPEAT_SPAN     10  // number of key repeat interval frames

/*---------------------------------------------------------------------------*
structure definitions
 *---------------------------------------------------------------------------*/

// key input data
typedef struct KeyInfo
{
    u16 cnt;    // unprocessed input values
    u16 trg;    // push trigger input
    u16 up;     // release trigger input
    u16 rep;    // push and hold repeat input
} KeyInfo;

/*---------------------------------------------------------------------------*
internal variable definitions
 *---------------------------------------------------------------------------*/

// key input
static KeyInfo  gKey;

/*---------------------------------------------------------------------------*
internal function prototype
 *---------------------------------------------------------------------------*/
static void     ProcMain(void);
static void     VBlankIntr(void);
static void     InitInterrupts(void);
static void     InitHeap(void);
static void     ReadKey(KeyInfo* pKey);

/*---------------------------------------------------------------------------*
Name: NitroMain

Description: Initialization and main loop.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
void NitroMain(void)
{
// the different types of initialization
    OS_Init();
    OS_InitTick();  // needed for use by the OS_Sleep functions
    OS_InitAlarm(); // needed for use by the OS_Sleep functions
    GX_Init();
    GX_DispOff();
    GXS_DispOff();

// initialization of memory allocation
    InitHeap();

// initializes display settings
    InitScreen();

// initializes interrupt settings
    InitInterrupts();

// LCD display start
    GX_DispOn();
    GXS_DispOn();

// debug string output
    OS_TPrintf("\"wcm-1\" demo started.\n");
    PrintSubScreen("\"wcm-1\" demo started.\n");

// empty call for getting key input data (strategy for pressing A button in IPL)
    ReadKey(&gKey);

    while (TRUE)
    {
// get key input data
        ReadKey(&gKey);

// clear the main screen
        ClearMainScreen();

        ProcMain();

// waiting for Vblank (thread support)
        OS_WaitVBlankIntr();
    }
}

/*---------------------------------------------------------------------------*
Name: ProcMain

Description: Displays the WCM library connection management status and performs
process switching based on input. Called from the main loop
for each picture frame.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void ProcMain(void)
{
    switch (WCM_GetPhase())
    {
    case WCM_PHASE_NULL:
// initial state or state when WCM control thread is complete
        ((u16*)HW_PLTT)[0] = 0x7fff;
        PutMainScreen(11, 10, 0x0, "Now offline");
        PutMainScreen(4, 12, (u8) (((OS_GetVBlankCount() / 30) % 2) + 7), "Push any key to connect");
        if (gKey.trg | gKey.rep)
        {
            SetWcmScanType(WCM_OPTION_SCANTYPE_PASSIVE);
            InitWcmControl();
            PrintSubScreen("InitWcmControl()\n");
        }
        break;

    case WCM_PHASE_DCF:
// connected state
        ((u16*)HW_PLTT)[0] = 0x7ff0;
        PutMainScreen(11, 10, 0x0, "Now online");
        PutMainScreen(3, 12, (u8) (((OS_GetVBlankCount() / 30) % 2) + 7), "Push any key to disconnect");
        if (gKey.trg | gKey.rep)
        {
            TerminateWcmControl();
            PrintSubScreen("TerminateWcmControl()\n");
        }
        break;

    case WCM_PHASE_FATAL_ERROR:
// unrecoverable fatal error
        ((u16*)HW_PLTT)[0] = 0x001f;
        PutMainScreen(0, 7, 0xf, "================================");
        PutMainScreen(7, 10, 0xf, "WCM FATAL ERROR !");
        PutMainScreen(2, 12, 0xf, "Turn off power supply please.");
        PutMainScreen(0, 15, 0xf, "================================");
        OS_WaitVBlankIntr();
        OS_Terminate();
        break;

    default:
// Processing
        ((u16*)HW_PLTT)[0] = 0x6318;
        PutMainScreen(10, 10, 0x0, "Now working...");
        PutMainScreen(6, 12, (u8) (((OS_GetVBlankCount() / 30) % 2) + 7), "Push any key to stop");
        if (gKey.trg | gKey.rep)
        {
            TerminateWcmControl();
            PrintSubScreen("TerminateWcmControl()\n");
        }
    }
}

/*===========================================================================*/

/*---------------------------------------------------------------------------*
Name: VBlankIntr

Description: VBlank interrupt handler.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void VBlankIntr(void)
{
// updates text display
    UpdateScreen();

// sets the IRQ check flag
    OS_SetIrqCheckFlag(OS_IE_V_BLANK);
}

/*---------------------------------------------------------------------------*
Name: InitInterrupts

Description: Initializes the interrupt settings.
Allows VBlank interrupts and sets the interrupt handler.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void InitInterrupts(void)
{
// VBlank interrupt settings
    OS_SetIrqFunction(OS_IE_V_BLANK, VBlankIntr);
    (void)OS_EnableIrqMask(OS_IE_V_BLANK);
    (void)GX_VBlankIntr(TRUE);

// interrupt permissions
    (void)OS_EnableIrq();
    (void)OS_EnableInterrupts();
}

/*---------------------------------------------------------------------------*
Name: InitHeap

Description: Initializes the memory allocation system within the main memory arena.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void InitHeap(void)
{
    void*           tempLo;
    OSHeapHandle    hh;

// creates a heap in the main memory arena
    tempLo = OS_InitAlloc(OS_ARENA_MAIN, OS_GetMainArenaLo(), OS_GetMainArenaHi(), 1);
    OS_SetArenaLo(OS_ARENA_MAIN, tempLo);
    hh = OS_CreateHeap(OS_ARENA_MAIN, OS_GetMainArenaLo(), OS_GetMainArenaHi());
    if (hh < 0)
    {
// abnormal end when heap creation fails
        OS_Panic("ARM9: Fail to create heap...\n");
    }
    (void)OS_SetCurrentHeap(OS_ARENA_MAIN, hh);
}

/*---------------------------------------------------------------------------*
Name: ReadKey

Description: Gets key input data, and edits the input data structure.
Detects push, release and push and hold repeat triggers.

Arguments: pKey  - Specifies the key input structure to be changed.

Returns: None.
 *---------------------------------------------------------------------------*/
static void ReadKey(KeyInfo* pKey)
{
    static u16  repeat_count[12];
    int         i;
    u16         r;

    r = PAD_Read();
    pKey->trg = 0x0000;
    pKey->up = 0x0000;
    pKey->rep = 0x0000;

    for (i = 0; i < 12; i++)
    {
        if (r & (0x0001 << i))
        {
            if (!(pKey->cnt & (0x0001 << i)))
            {
                pKey->trg |= (0x0001 << i);     // push trigger
                repeat_count[i] = 1;
            }
            else
            {
                if (repeat_count[i] > KEY_REPEAT_START)
                {
                    pKey->rep |= (0x0001 << i); // push and hold repeat
                    repeat_count[i] = (u16) (KEY_REPEAT_START - KEY_REPEAT_SPAN);
                }
                else
                {
                    repeat_count[i]++;
                }
            }
        }
        else
        {
            if (pKey->cnt & (0x0001 << i))
            {
                pKey->up |= (0x0001 << i);      // release trigger
            }
        }
    }

    pKey->cnt = r;  // unprocessed key input
}

/*---------------------------------------------------------------------------*
End of file
 *---------------------------------------------------------------------------*/
