/*---------------------------------------------------------------------------*
Project: NitroWiFi - WCM - demos - wcm-list
File: main.c

Copyright 2005.2006 Nintendo.  All rights reserved.

These coded instructions, statements, and computer programs contain
proprietary information of Nintendo of America Inc. and/or Nintendo
Company Ltd., and are protected by Federal copyright law.  They may
not be disclosed to third parties or copied or duplicated in any form,
in whole or in part, without the prior written consent of Nintendo.

$Log: main.c,v $
Revision 1.7  03/10/2006 09:22:43  kitase_hirotake
  INDENT SOURCE

Revision 1.6  2005/11/02 02:41:33  terui
Fix comment.

Revision 1.5  10/25/2005 04:19:19  terui
SVC_WaitVBlankIntr -> OS_WaitVBlankIntr

Revision 1.4  2005/08/29 06:39:36  yosizaki
add SetWcmManualConnect().

Revision 1.3  07/25/2005 07:48:46  adachi_hiroaki
added a list reset button

Revision 1.2  07/22/2005 01:07:40  adachi_hiroaki
made it such that it will only move to selection move if the default BSSID is ANY

Revision 1.1  07/21/2005 08:21:06  adachi_hiroaki
new additions


$NoKeywords: $
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*
Sample that uses the WCM library to search for general wireless access points (AP),
    lists the points, and then connects.
 *---------------------------------------------------------------------------*/
#include <nitro.h>
#include <nitroWiFi.h>
#include "screen.h"
#include "wcm_control.h"

/*---------------------------------------------------------------------------*
constant definitions
 *---------------------------------------------------------------------------*/
#define KEY_REPEAT_START    25  // number of frames until key repeat starts
#define KEY_REPEAT_SPAN     10  // number of key repeat interval frames

/*---------------------------------------------------------------------------*
structure definitions
 *---------------------------------------------------------------------------*/

// key input data
typedef struct KeyInfo
{
    u16 cnt;    // unprocessed input values
    u16 trg;    // push trigger input
    u16 up;     // release trigger input
    u16 rep;    // push and hold repeat input
} KeyInfo;

/*---------------------------------------------------------------------------*
internal variable definitions
 *---------------------------------------------------------------------------*/

// key input
static KeyInfo  gKey;

//default connection target BSSID
static u8       gBssidDefault[] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
static int      gSelectedAP = 0;
static BOOL     gConnectAuto = FALSE;

/*---------------------------------------------------------------------------*
internal function prototype
 *---------------------------------------------------------------------------*/
static void     ProcMain(void);
static void     SelectAP(void);
static const WMBssDesc*     GetBssDesc(const u8* bssid);
static void VBlankIntr(void);
static void InitInterrupts(void);
static void InitHeap(void);
static void ReadKey(KeyInfo* pKey);

/*---------------------------------------------------------------------------*
Name: NitroMain

Description: Initialization and main loop.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
void NitroMain(void)
{
// the different types of initialization
    OS_Init();
    OS_InitTick();  // needed for use by the OS_Sleep functions
    OS_InitAlarm(); // needed for use by the OS_Sleep functions
    GX_Init();
    GX_DispOff();
    GXS_DispOff();

// initialization of memory allocation
    InitHeap();

// initializes display settings
    InitScreen();

// initializes interrupt settings
    InitInterrupts();

// LCD display start
    GX_DispOn();
    GXS_DispOn();

// debug string output
    OS_TPrintf("\"wcm-list\" demo started.\n");
    PrintSubScreen("\"wcm-list\" demo started.\n");

// empty call for getting key input data (strategy for pressing A button in IPL)
    ReadKey(&gKey);

    while (TRUE)
    {
// get key input data
        ReadKey(&gKey);

// clear the main screen
        ClearMainScreen();

        ProcMain();

// waiting for Vblank (thread support)
        OS_WaitVBlankIntr();
    }
}

/*---------------------------------------------------------------------------*
Name: ProcMain

Description: Displays the WCM library connection management status and performs
process switching based on input. Called from the main loop
for each picture frame.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void ProcMain(void)
{
    switch (WCM_GetPhase())
    {
    case WCM_PHASE_NULL:
// initial state or state when WCM control thread is complete
        ((u16*)HW_PLTT)[0] = 0x7fff;
        PutMainScreen(11, 10, 0x0, "Now offline");
        PutMainScreen(4, 12, (u8) (((OS_GetVBlankCount() / 30) % 2) + 7), "Push any key to search AP");
        if (gKey.trg | gKey.rep)
        {
            SetWcmManualConnect(TRUE);
            InitWcmControl();
        }

        if (!WCM_CompareBssID(WCM_BSSID_ANY, gBssidDefault))
        {
            PutMainScreen(5, 23, 0x0, "Start button: auto connect");
            if (gKey.trg & PAD_BUTTON_START)
            {
                gConnectAuto = TRUE;
            }
            else
            {
                gConnectAuto = FALSE;
            }
        }
        break;

    case WCM_PHASE_SEARCH:
// AP while searching
        if (gConnectAuto)
        {
// auto-connect mode
            PutMainScreen(11, 10, 0x0, "Now online");
            PutMainScreen(4, 11, 0x0, "connecting to default AP");
            PutMainScreen(3, 12, (u8) (((OS_GetVBlankCount() / 30) % 2) + 7), "Push any key to select AP");
            if (gKey.trg | gKey.rep)
            {
// to AP selection mode if a key is pressed
                gConnectAuto = FALSE;
            }
            else
            {
//search for the AP with the default BSSID
                const WMBssDesc*    pbd = GetBssDesc(gBssidDefault);
                if (pbd)
                {
//if found, connect
                    ConnectAP(pbd);
                }
            }
        }
        else
        {
// AP selection mode
            SelectAP();
        }
        break;

    case WCM_PHASE_DCF:
// connected state
        ((u16*)HW_PLTT)[0] = 0x7ff0;
        PutMainScreen(11, 10, 0x0, "Now online");
        PutMainScreen(3, 12, (u8) (((OS_GetVBlankCount() / 30) % 2) + 7), "Push any key to disconnect");
        if (gKey.trg | gKey.rep)
        {
            TerminateWcmControl();
        }
        break;

    case WCM_PHASE_FATAL_ERROR:
// unrecoverable fatal error
        ((u16*)HW_PLTT)[0] = 0x001f;
        PutMainScreen(0, 7, 0xf, "================================");
        PutMainScreen(7, 10, 0xf, "WCM FATAL ERROR !");
        PutMainScreen(2, 12, 0xf, "Turn off power supply please.");
        PutMainScreen(0, 15, 0xf, "================================");
        OS_WaitVBlankIntr();
        OS_Terminate();
        break;

    default:
// Processing
        ((u16*)HW_PLTT)[0] = 0x6318;
        PutMainScreen(10, 10, 0x0, "Now working...");
        PutMainScreen(6, 12, (u8) (((OS_GetVBlankCount() / 30) % 2) + 7), "Push any key to stop");
        if (gKey.trg | gKey.rep)
        {
            TerminateWcmControl();
        }
    }
}

/*---------------------------------------------------------------------------*
Name: SelectAP

Description: displays a list of usable access points to choose from

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void SelectAP(void)
{
    int     nIndex = 0;
    int     nAPCount = 0;
    char    buffer[256];
    s32     nLevel;
    BOOL    bDefault = FALSE;

    PutMainScreen(0, 0, 0xf, " bssid      + ssid/link");
    PutMainScreen(0, 23, 0xf, "A:connect, Sel:set def, St:clear");

    (void)WCM_LockApList(WCM_APLIST_LOCK);
    nAPCount = WCM_CountApList();
    if (nAPCount == 0)
    {
//no display or selection if a single AP cannot be found
        (void)WCM_LockApList(WCM_APLIST_UNLOCK);
        return;
    }

//move the selection cursor
    if (gKey.trg & PAD_KEY_DOWN)
    {
        ++gSelectedAP;
    }

    if (gKey.trg & PAD_KEY_UP)
    {
        --gSelectedAP;
    }

    gSelectedAP = MATH_CLAMP(gSelectedAP, 0, nAPCount - 1);
    {
        WMBssDesc*  pbdSelected = WCM_PointApList(gSelectedAP);
        if (pbdSelected)
        {
            if (gKey.trg & PAD_BUTTON_A)
            {
                ConnectAP(pbdSelected);
            }

            if (gKey.trg & PAD_BUTTON_SELECT)
            {
                MI_CpuCopy8(pbdSelected->bssid, gBssidDefault, WM_SIZE_BSSID);
            }
        }
    }

    for (nIndex = 0; nIndex < nAPCount; ++nIndex)
    {
        WMBssDesc*  pbd = WCM_PointApList(nIndex);
        bDefault = WCM_CompareBssID(gBssidDefault, pbd->bssid);
        nLevel = WCM_PointApListLinkLevel(nIndex);
        (void)OS_SPrintf(buffer, "%02X%02X%02X%02X%02X%02X%s%s/%d", pbd->bssid[0], pbd->bssid[1], pbd->bssid[2],
                         pbd->bssid[3], pbd->bssid[4], pbd->bssid[5], bDefault ? "*" : " ", pbd->ssid, nLevel);
        PutMainScreen(0, 1 + nIndex, (u8) (nIndex == gSelectedAP ? 0x1 : 0xf), buffer);
    }
    (void)WCM_LockApList(WCM_APLIST_UNLOCK);

    if (gKey.trg & PAD_BUTTON_START)
    {
        WCM_ClearApList();
    }

}

/*---------------------------------------------------------------------------*
Name: GetBssDesc

Description: Get the BssDesc if an AP with the specified BSSID is found.

Arguments: BSSID for the AP whose BssDesc is wanted

Returns: pointer to the BssDesc in the found AP
 *---------------------------------------------------------------------------*/
static const WMBssDesc* GetBssDesc(const u8* bssid)
{
    int nAPCount = 0;
    int nIndex = 0;

    (void)WCM_LockApList(WCM_APLIST_LOCK);
    nAPCount = WCM_CountApList();
    for (nIndex = 0; nIndex < nAPCount; ++nIndex)
    {
        WMBssDesc*  pbd = WCM_PointApList(nIndex);
        if (WCM_CompareBssID((u8*)bssid, pbd->bssid))
        {
            (void)WCM_LockApList(WCM_APLIST_UNLOCK);
            return pbd;
        }
    }
    (void)WCM_LockApList(WCM_APLIST_UNLOCK);
    return NULL;
}

/*===========================================================================*/

/*---------------------------------------------------------------------------*
Name: VBlankIntr

Description: VBlank interrupt handler.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void VBlankIntr(void)
{
// updates text display
    UpdateScreen();

// sets the IRQ check flag
    OS_SetIrqCheckFlag(OS_IE_V_BLANK);
}

/*---------------------------------------------------------------------------*
Name: InitInterrupts

Description: Initializes the interrupt settings.
Allows VBlank interrupts and sets the interrupt handler.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void InitInterrupts(void)
{
// VBlank interrupt settings
    OS_SetIrqFunction(OS_IE_V_BLANK, VBlankIntr);
    (void)OS_EnableIrqMask(OS_IE_V_BLANK);
    (void)GX_VBlankIntr(TRUE);

// interrupt permissions
    (void)OS_EnableIrq();
    (void)OS_EnableInterrupts();
}

/*---------------------------------------------------------------------------*
Name: InitHeap

Description: Initializes the memory allocation system within the main memory arena.

Arguments: None.

Returns: None.
 *---------------------------------------------------------------------------*/
static void InitHeap(void)
{
    void*           tempLo;
    OSHeapHandle    hh;

// creates a heap in the main memory arena
    tempLo = OS_InitAlloc(OS_ARENA_MAIN, OS_GetMainArenaLo(), OS_GetMainArenaHi(), 1);
    OS_SetArenaLo(OS_ARENA_MAIN, tempLo);
    hh = OS_CreateHeap(OS_ARENA_MAIN, OS_GetMainArenaLo(), OS_GetMainArenaHi());
    if (hh < 0)
    {
// abnormal end when heap creation fails
        OS_Panic("ARM9: Fail to create heap...\n");
    }
    (void)OS_SetCurrentHeap(OS_ARENA_MAIN, hh);
}

/*---------------------------------------------------------------------------*
Name: ReadKey

Description: Gets key input data, and edits the input data structure.
Detects push, release and push and hold repeat triggers.

Arguments: pKey  - Specifies the key input structure to be changed.

Returns: None.
 *---------------------------------------------------------------------------*/
static void ReadKey(KeyInfo* pKey)
{
    static u16  repeat_count[12];
    int         i;
    u16         r;

    r = PAD_Read();
    pKey->trg = 0x0000;
    pKey->up = 0x0000;
    pKey->rep = 0x0000;

    for (i = 0; i < 12; i++)
    {
        if (r & (0x0001 << i))
        {
            if (!(pKey->cnt & (0x0001 << i)))
            {
                pKey->trg |= (0x0001 << i);     // push trigger
                repeat_count[i] = 1;
            }
            else
            {
                if (repeat_count[i] > KEY_REPEAT_START)
                {
                    pKey->rep |= (0x0001 << i); // push and hold repeat
                    repeat_count[i] = (u16) (KEY_REPEAT_START - KEY_REPEAT_SPAN);
                }
                else
                {
                    repeat_count[i]++;
                }
            }
        }
        else
        {
            if (pKey->cnt & (0x0001 << i))
            {
                pKey->up |= (0x0001 << i);      // release trigger
            }
        }
    }

    pKey->cnt = r;  // unprocessed key input
}

/*---------------------------------------------------------------------------*
End of file
 *---------------------------------------------------------------------------*/
